#include "includes/automaton.hpp"
#include "includes/temps.h"


Automaton::Automaton(config & cfg)
{
int i;
stringstream out;
//build the different modules
  if (cfg.OUTHQI !=0) hqi=new HQI();else hqi=NULL;	
  if (cfg.OUTBLUE !=0) blue=new Blue(cfg.OUTBLUE,cfg.OUTBLUE_CARD,"1"); else blue=NULL;
  if (cfg.OUTBLUE2 !=0) blue2=new Blue(cfg.OUTBLUE2,cfg.OUTBLUE2_CARD,"2"); else blue2=NULL;
  if (cfg.OUTPUMP !=0) pump=new Pump(cfg.OUTPUMP,cfg.OUTPUMP_CARD,"1"); else pump=NULL;
  if (cfg.OUTPUMP2 !=0) pump2=new Pump(cfg.OUTPUMP2,cfg.OUTPUMP2_CARD,"2"); else pump2=NULL;
  if (cfg.OUTTEMP !=0) temperature=new Temperature(); else temperature=NULL;
  if (cfg.OUTSKIMMER !=0) skimmer=new Skimmer(); else skimmer=NULL;
for(i=0;i<4;i++)
	{
  	if (cfg.OUTPERI[i] !=0) peristatic[i]=new Peristatic(out.str()); else peristatic[i]=NULL;
	}
  if (cfg.OUTRAH !=0) rah=new Rah(); else rah=NULL;
  if (cfg.OUTOSMOLATION !=0) osmolation=new Osmolation(); else osmolation=NULL;
  if (cfg.OUTVENT !=0) ventilator=new Ventilator(); else ventilator=NULL;
  if (cfg.PH !=0)  ph=new phsonde(); else ph=NULL;
  if (cfg.OUTPREMONT !=0)  remont=new Premont(); else remont=NULL;

//  init();
getstart=0;
}

Automaton::~Automaton()
{
int i;
//delete all modules
  delete hqi;
  delete blue;
  delete blue2;
  delete skimmer;
  delete temperature;
  delete pump;
  delete pump2;
for (i=0;i<4;i++)  
	if (peristatic[i]!=NULL) delete peristatic[i];
  delete rah;
  delete osmolation;
  delete ventilator;
  delete ph;
  delete remont; 
}

void Automaton::restart (connexion & c)
{
  int n;
  
//restart the automaton without saving BD
  if (getstart==1)
    {
      // normal restart
      if (DEBUG==1) cout << "restarting automaton" << endl;
      c.insert_log("automate redemarré");
      n=fork();
      if (n==0) {
	if (MODE=="LINUX")
	execl("/usr/bin/sudo","sudo automate","./automate",NULL);
	if (MODE=="OPENWRT")
	execl("/bin/automate","automate",NULL);	
		}
      
    }
  
//restart the automaton with reinit of BD
  if (getstart==2)
    {
      if (DEBUG==1) cout << "restarting automaton with fresh bd" << endl;
      c.insert_log("automate redemarré avec nouvelle bd");
      n=fork();
      if (n==0) {
	if (MODE=="LINUX")
	execl("/usr/bin/sudo","sudo automate","./automate","init-bd",NULL);
	if (MODE=="OPENWRT")
	execl("/bin/automate","automate","init-bd",NULL);	
		}
    }
  
//stop the automaton
  if (getstart==3)
    {
      if (DEBUG==1) cout << "stopping automaton" << endl;
      c.insert_log("automate stoppé");
    }
}

//getting input states

void Automaton::getInputState(K8055 &k,config &cfg)
{
short h=0;
if (cfg.HIGH!=0) h=k.read(cfg.HIGH,cfg.HIGH_CARD);

  if (cfg.OUTPUMP != 0 && cfg.STOP_PUMP!=0) pump->setStop(k.read(cfg.STOP_PUMP,cfg.STOP_PUMP_CARD));

  if (cfg.OUTPUMP2 != 0 && cfg.STOP_PUMP!=0) pump2->setStop(k.read(cfg.STOP_PUMP,cfg.STOP_PUMP_CARD));

  if (cfg.OUTSKIMMER != 0 && cfg.STOP_SKIMMER !=0) 
	skimmer->setStop(k.read(cfg.STOP_SKIMMER,cfg.STOP_SKIMMER_CARD));
  if (cfg.OUTSKIMMER != 0 && cfg.HIGH!=0)	
		skimmer->setCaptor_high(h);

 if (cfg.OUTOSMOLATION != 0 )
	{
	if (cfg.OSMOLATION!=0)
		osmolation->setIni(k.read(cfg.OSMOLATION,cfg.OSMOLATION_CARD));
	if (cfg.RESERV!=0)
  		osmolation->setEmpty_reserv(k.read(cfg.RESERV,cfg.RESERV_CARD));
	if (cfg.HIGH!=0)  	
		osmolation->setCaptor_high(h);
	}
if (cfg.OUTPREMONT !=0)
	remont->setCaptor(k.read(cfg.STOP_PREMONT,cfg.STOP_PREMONT_CARD));
}
//return preristatic obj i
Peristatic * Automaton::getPeristatic(int i)
{
 if (i<4 && i>=0) return peristatic[i];
else return NULL;
}


