
#include "stdafx.h"
#include "cphidgetmanager.h"
#include "cphidget.h"
#include "cthread.h"
#include "cusb.h"
#include "cphidgetlist.h"
#include "csocket.h"

/* A list of local Phidget Managers */
CPhidgetManagerListHandle localPhidgetManagers = NULL;
int ActivePhidgetManagers = 0;

/* Protects localPhidgetManagers */
int managerLockInitialized = PFALSE;
CThread_mutex_t managerLock;

int CPhidgetManager_areEqual(void *arg1, void *arg2)
{
	if(arg1 == arg2) return 1;
	return 0;
}

void CPhidgetManager_free(void *arg)
{
	CPhidgetManagerHandle phidm = (CPhidgetManagerHandle)arg;
	
	if(!phidm) return;
	free(phidm);
	return;
}

int CCONV CPhidgetManager_set_OnError_Handler(CPhidgetManagerHandle phidm, 
	int(CCONV *fptr)(CPhidgetManagerHandle, void *, int, const char *), void *userPtr)
{
	if (!phidm)
		return EPHIDGET_INVALIDARG;
	phidm->fptrError = fptr;
	phidm->fptrErrorptr = userPtr;
	return EPHIDGET_OK;
}

/* The internal event handlers for PhidgetManager - these should be called directly from the usb functions...
 * these run in the context of the central thread
 * if AddDevice returns EPHIDGET_DUPLICATE, the user callback is not run (good)
 */
int CPhidgetAttachEvent(CPhidgetHandle phid) {
	int result = 0;
	CPhidgetList *trav = 0;
	CPhidgetManagerList *trav2 = 0;
	if (!phid)
		return EPHIDGET_INVALIDARG;
	result = CList_addToList((CListHandle *)&AttachedDevices, phid, CPhidget_areEqual);
	if(result == EPHIDGET_DUPLICATE)
	{
		return EPHIDGET_OK;
	}
	else if(result) 
	{
		return result;
	}
	
	for(trav2 = localPhidgetManagers; trav2; trav2 = trav2->next)
	{
		if (trav2->phidm->fptrAttachChange && trav2->phidm->state == PHIDGETMANAGER_ACTIVE)
			trav2->phidm->fptrAttachChange((CPhidgetHandle)phid, trav2->phidm->fptrAttachChangeptr);
	}

	for (trav=ActiveDevices; trav; trav = trav->next)
	{
		if(CPhidget_areEqual(trav->phid, phid)
			&& !CPhidget_isAttached(trav->phid)
			&& CList_findInList((CListHandle)ActiveDevices, phid, CPhidget_areEqual, NULL) != EPHIDGET_OK)
		{
			if(trav->phid->specificDevice == 0)
			{
				trav->phid->specificDevice = 2;
				trav->phid->serialNumber = phid->serialNumber;
			}

			if ((result = CUSBOpenHandle(trav->phid)) !=
			    EPHIDGET_OK) {
				WARN("unable to open active device");
				return result;
			}

			CThread_mutex_lock(&trav->phid->lock);
			trav->phid->status = PHIDGET_ATTACHED;
			trav->phid->fptrInit((CPhidgetHandle)trav->phid, NULL, 0);
			
			/* Start read thread */
			if (CThread_create(&trav->phid->readThread,
			    ReadThreadFunction, trav->phid)) {
				WARN("unable to create read thread");
				trav->phid->status = PHIDGET_NOTATTACHED;
				CThread_mutex_unlock(&trav->phid->lock);
				return EPHIDGET_UNEXPECTED;
			}
			if (CThread_create(&trav->phid->writeThread,
			    WriteThreadFunction, trav->phid)) {
				WARN("unable to create write thread");
				trav->phid->status = PHIDGET_NOTATTACHED;
				CThread_mutex_unlock(&trav->phid->lock);
				return EPHIDGET_UNEXPECTED;
			}
			
			CThread_mutex_unlock(&trav->phid->lock);
			
			if (trav->phid->fptrAttach)
				trav->phid->fptrAttach(trav->phid, trav->phid->fptrAttachptr);
			break;
		}
	}
	return EPHIDGET_OK;
}

int CPhidgetDetachEvent(CPhidgetHandle phid) {
	int result = 0;
	CPhidgetList *trav = 0;
	CPhidgetManagerList *trav2 = 0;
	if(!phid) return EPHIDGET_INVALIDARG;
	phid->status = PHIDGET_NOTATTACHED;
	for(trav2 = localPhidgetManagers; trav2; trav2 = trav2->next)
	{
		if (trav2->phidm->fptrDetachChange && trav2->phidm->state == PHIDGETMANAGER_ACTIVE)
			trav2->phidm->fptrDetachChange((CPhidgetHandle)phid, trav2->phidm->fptrDetachChangeptr);
	}

	for (trav=ActiveDevices; trav; trav = trav->next)
	{		
		if(CPhidget_areEqual(phid, trav->phid)
			&& CPhidget_isAttached(trav->phid))
		{
			if(trav->phid->specificDevice == 2)
				trav->phid->specificDevice = 0;
			trav->phid->deviceIDSpec = 0;

			CUSBCloseHandle(trav->phid);
			if (trav->phid->fptrDetach)
				trav->phid->fptrDetach((CPhidgetHandle)trav->phid, trav->phid->fptrDetachptr);
			break;
		}

	}

	result = CList_removeFromList((CListHandle *)&AttachedDevices, phid, CPhidget_areEqual, TRUE, CPhidget_free);

	return result;
}

int CPhidgetManager_poll()
{
	CPhidgetList *curList = 0, *detachList = 0;
	CPhidgetList *trav = 0;
			
	if(!managerLockInitialized)
	{
		CThread_mutex_init(&managerLock);
		managerLockInitialized = PTRUE;
	}
	
	CThread_mutex_lock(&managerLock);
	CThread_mutex_lock(&phidgetLock);
	CUSBBuildList(&curList);

	for (trav=AttachedDevices; trav; trav = trav->next)
	{		
		if(CList_findInList((CListHandle)curList, trav->phid, CPhidget_areEqual, NULL) == EPHIDGET_NOTFOUND)
		{
			CList_addToList((CListHandle *)&detachList, trav->phid, CPhidget_areEqual);
		}
	}
	for (trav=curList; trav; trav = trav->next)
	{
		if(CList_findInList((CListHandle)AttachedDevices, trav->phid, CPhidget_areEqual, NULL) ==
		    EPHIDGET_NOTFOUND)
		{
			CPhidgetAttachEvent(trav->phid);
		}
	}
	for (trav=detachList; trav; trav = trav->next)
	{
		CPhidgetDetachEvent(trav->phid);
	}
	CList_emptyList((CListHandle *)&detachList, FALSE, NULL);
	CList_emptyList((CListHandle *)&curList, FALSE, NULL);

	CThread_mutex_unlock(&phidgetLock);
	CThread_mutex_unlock(&managerLock);
	return EPHIDGET_OK;
}

int CCONV CPhidgetManager_create(CPhidgetManagerHandle *phidm)
{
	CPhidgetManagerHandle phidmtemp = 0;
	
	if(!phidm)
		return EPHIDGET_INVALIDARG;
	if(!(phidmtemp = (CPhidgetManagerHandle)malloc(sizeof(CPhidgetManager))))
		return EPHIDGET_NOMEMORY;
	ZEROMEM(phidmtemp, sizeof(CPhidgetManager));
	
	phidmtemp->state = PHIDGETMANAGER_INACTIVE;
	
	if(!managerLockInitialized)
	{
		CThread_mutex_init(&managerLock);
		managerLockInitialized = PTRUE;
	}

	CThread_mutex_init(&phidmtemp->lock);
	
	*phidm = phidmtemp;
	return EPHIDGET_OK;
}

int CCONV CPhidgetManager_close(CPhidgetManagerHandle phidm)
{
	if(!phidm)
		return EPHIDGET_INVALIDARG;

	if(phidm->state == PHIDGETMANAGER_ACTIVE)
	{
		phidm->state = PHIDGETMANAGER_INACTIVE;
		CPhidget_setStatus((CPhidgetHandle)phidm, PHIDGET_NOTATTACHED);
		if(phidm->networkInfo)
		{
			CThread_mutex_lock(&serverLock);
			unregisterRemoteManager(phidm);
			CThread_mutex_unlock(&serverLock);
		}
		else
		{
			CThread_mutex_lock(&managerLock);
			ActivePhidgetManagers--;
			CList_removeFromList((CListHandle *)&localPhidgetManagers, phidm, CPhidgetManager_areEqual, PFALSE, NULL);
			CThread_mutex_unlock(&managerLock);
		}
	}
	return EPHIDGET_OK;
}

int CCONV CPhidgetManager_delete(CPhidgetManagerHandle phidm)
{
	if(!phidm)
		return EPHIDGET_INVALIDARG;
	if(phidm->networkInfo)
	{
		CList_emptyList((CListHandle *)phidm->AttachedPhidgets, PTRUE, CPhidget_free);
	}
	CPhidgetManager_free(phidm);
	return EPHIDGET_OK;
}

int CCONV CPhidgetManager_open(CPhidgetManagerHandle phidm)
{
	int result = EPHIDGET_OK;
	
	if(!phidm)
		return EPHIDGET_INVALIDARG;
	
	if(!phidgetLockInitialized)
	{
		CThread_mutex_init(&phidgetLock);
		phidgetLockInitialized = PTRUE;
	}

	if(phidm->state == PHIDGETMANAGER_INACTIVE)
	{
		CThread_mutex_lock(&managerLock);
	
		CList_addToList((CListHandle *)&localPhidgetManagers, phidm, CPhidgetManager_areEqual);
		
		phidm->state = PHIDGETMANAGER_ACTIVE;
		
		CPhidget_setStatus((CPhidgetHandle)phidm, PHIDGET_ATTACHED);

		phidm->AttachedPhidgets = AttachedDevices;
		
		CThread_mutex_lock(&phidgetLock);
		result = CUSBBuildList(&AttachedDevices);
		if(result) 
		{
			CThread_mutex_unlock(&phidgetLock);
			CThread_mutex_unlock(&managerLock);
			return result;
		}
		
		if (phidm->fptrAttachChange)
		{
			CPhidgetList *trav = 0;
			
			for (trav=AttachedDevices; trav; trav = trav->next) {
				phidm->fptrAttachChange((CPhidgetHandle)(trav->phid), phidm->fptrAttachChangeptr);
			}
		}
		ActivePhidgetManagers++;
		
		CThread_mutex_unlock(&phidgetLock);
		CThread_mutex_unlock(&managerLock);
		
		result = StartCentralThread();
	}
	return result;
}

int CCONV CPhidgetManager_set_OnAttach_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *userPtr)
{
	if(!phidm)
		return EPHIDGET_INVALIDARG;
	phidm->fptrAttachChange = fptr; 
	phidm->fptrAttachChangeptr = userPtr; 
	return EPHIDGET_OK; 
}
int CCONV CPhidgetManager_set_OnDetach_Handler(CPhidgetManagerHandle phidm, int (CCONV *fptr)(CPhidgetHandle phid, void *userPtr), void *userPtr)
{
	if(!phidm)
		return EPHIDGET_INVALIDARG;
	phidm->fptrDetachChange = fptr; 
	phidm->fptrDetachChangeptr = userPtr; 
	return EPHIDGET_OK; 
}

int CCONV CPhidgetManager_getAttachedDevices(CPhidgetManagerHandle phidm, CPhidgetHandle *phidArray[],
    int *count)
{
	CPhidgetList *trav = 0;
	int i = 0;

	if (!phidArray || !count)
		return EPHIDGET_INVALIDARG;
	if(!phidm)
		return EPHIDGET_INVALIDARG;

	*count = 0;
	if(phidm->networkInfo)
	{		
		for (trav=phidm->AttachedPhidgets; trav; trav = trav->next)
			*count++;
		
		*phidArray = (CPhidgetHandle *)malloc(sizeof(**phidArray) * *count);
		if (!*phidArray)
			return EPHIDGET_NOMEMORY;
		ZEROMEM(*phidArray, sizeof(**phidArray) * *count);

		for (trav=phidm->AttachedPhidgets, i=0; trav; trav = trav->next, i++)
			(*phidArray)[i] = trav->phid;
	}
	else
	{
		CThread_mutex_lock(&phidgetLock);
		for (trav=AttachedDevices; trav; trav = trav->next)
			*count++;
		
		*phidArray = (CPhidgetHandle *)malloc(sizeof(**phidArray) * *count);
		if (!*phidArray)
			return EPHIDGET_NOMEMORY;
		ZEROMEM(*phidArray, sizeof(**phidArray) * *count);

		for (trav=AttachedDevices, i=0; trav; trav = trav->next, i++)
			(*phidArray)[i] = trav->phid;
		CThread_mutex_unlock(&phidgetLock);
	}
	return EPHIDGET_OK; 
}
