#include <assert.h>
#include "stdafx.h"
#include "cphidget.h"
#include "cphidgetmanager.h"
#include "cusb.h"
#include "cphidgetlist.h"

CThread_func_return_t CentralThreadFunction(CThread_func_arg_t arg);

static CThread CentralThread;

#ifdef _MACOSX
void macManagerThread(CFRunLoopTimerRef timer, void *Handle);
#endif

int
FindActiveDevices()
{
	CPhidgetList *trav = 0, *trav2;
	int result = 0;

	CThread_mutex_lock(&phidgetLock);
	for (trav=ActiveDevices; trav; trav = trav->next)
	{
		if(!CPhidget_isAttached(trav->phid))
		{
			for (trav2=AttachedDevices; trav2; trav2 = trav2->next)
			{		
				if(CPhidget_areEqual(trav->phid, trav2->phid)
					&& CList_findInList((CListHandle)ActiveDevices, trav2->phid, CPhidget_areEqual, NULL) != EPHIDGET_OK)
				{
					if(trav->phid->specificDevice == 0)
					{
						trav->phid->specificDevice = 2;
						trav->phid->serialNumber = trav2->phid->serialNumber;
					}

					if ((result = CUSBOpenHandle(trav->phid)) !=
						EPHIDGET_OK) {
						WARN("unable to open active device in findactivedevices: %d", result);
						CThread_mutex_unlock(&phidgetLock);
						return result;
					}
					
					CThread_mutex_lock(&trav->phid->lock);
					trav->phid->status = PHIDGET_ATTACHED;
					trav->phid->fptrInit((CPhidgetHandle)trav->phid, NULL, 0);
					
					//Start read thread
					if (CThread_create(&trav->phid->readThread,
						ReadThreadFunction, trav->phid)) {
						WARN("unable to create read thread");
						trav->phid->status = PHIDGET_NOTATTACHED;
						CThread_mutex_unlock(&trav->phid->lock);
						CThread_mutex_unlock(&phidgetLock);
						return EPHIDGET_UNEXPECTED;
					}
					if (CThread_create(&trav->phid->writeThread,
						WriteThreadFunction, trav->phid)) {
						WARN("unable to create write thread");
						trav->phid->status = PHIDGET_NOTATTACHED;
						CThread_mutex_unlock(&trav->phid->lock);
						CThread_mutex_unlock(&phidgetLock);
						return EPHIDGET_UNEXPECTED;
					}
					
					CThread_mutex_unlock(&trav->phid->lock);
					
					if (trav->phid->fptrAttach)
					{	
						//printf("Throwing read thread\n");
						trav->phid->fptrAttach(trav->phid,
							trav->phid->fptrAttachptr);
					}
					break;
				}
			}
		}
	}
	CThread_mutex_unlock(&phidgetLock);
	return EPHIDGET_OK;
}

int
StartCentralThread()
{

#ifdef _WINDOWS
	if (CentralThread.m_ThreadHandle) {
		int threadStatus = 0;
		int result = 0;
		result = GetExitCodeThread(CentralThread.m_ThreadHandle,
		    (LPDWORD)&threadStatus);
		if (result) {
			if (threadStatus != STILL_ACTIVE) {
				CloseHandle(CentralThread.m_ThreadHandle);
				CentralThread.m_ThreadHandle = 0;
			}
		}
	}
#endif

	if (!CentralThread.m_ThreadHandle)
		if (CThread_create(&CentralThread, CentralThreadFunction, 0))
			return EPHIDGET_UNEXPECTED;
	return EPHIDGET_OK;
}

/*
 * registers a device to recieve events and be polled by the central
 * thread This needs to start the central thread if it's not yet
 * running.
*/
int
RegisterLocalDevice(CPhidgetHandle phid)
{
	int result;

	if (!phid)
		return EPHIDGET_INVALIDARG;
	
	if(!phidgetLockInitialized)
	{
		CThread_mutex_init(&phidgetLock);
		phidgetLockInitialized = PTRUE;
	}
	CThread_mutex_lock(&phidgetLock);

	if(phid->specificDevice == TRUE)
		result = CList_addToList((CListHandle *)&ActiveDevices, phid, CPhidget_areEqual);
	else
		result = CList_addToList((CListHandle *)&ActiveDevices, phid, CPhidgetHandle_areEqual);

	if (result)
	{
		CThread_mutex_unlock(&phidgetLock);
		return result;
	}
	CThread_mutex_unlock(&phidgetLock);

	result = StartCentralThread();
	return result;
}


#ifdef _MACOSX
void macManagerThread(CFRunLoopTimerRef timer, void *Handle) {
	if(ActiveDevices || (ActivePhidgetManagers > 0)) {
		FindActiveDevices(); //this looks for attached active devices and opens them
	}
	return;
}
#endif

//The central thread should stop itself when there are no more active devices...? 
//Or we can stop it in unregisterlocaldevice

CThread_func_return_t CentralThreadFunction(CThread_func_arg_t lpdwParam)
{
	//Fist thing we need to do is set up a Phidget Manager and get the list of all phidgets already present
	//register to recieve notification in this thread's context

	//if the list isn't empty, first empty (free) it all...
	//result = CList_emptyList(&AttachedDevices, CPhidget_free); //I'm thinking that this is a bad idea
	//if (result) return result;

#ifdef _MACOSX
	CFRunLoopTimerRef timer;
	CPhidgetManager_setupNotifications(CFRunLoopGetCurrent());
	timer = CFRunLoopTimerCreate(kCFAllocatorDefault, 0, 0.250, 0, 0, macManagerThread, NULL);
	CFRunLoopAddTimer(CFRunLoopGetCurrent(), timer, kCFRunLoopDefaultMode);
	//start run loop
	CFRunLoopRun();
#else
	if(!phidgetLockInitialized)
	{
		CThread_mutex_init(&phidgetLock);
		phidgetLockInitialized = PTRUE;
	}
	//loop as long as there are active devices, or a phidget manager
	while(ActiveDevices || ActivePhidgetManagers) {
		CPhidgetManager_poll(); //this will update the list, as well as sending out attach and detach events
		FindActiveDevices(); //this looks for attached active devices and opens them
		SLEEP(250);
	}

	//if we actually get here, it means there are no active devices, and no phidgetmanagers, so free up the
	//last of the memory we are using, and exit the thread...
	CThread_mutex_lock(&phidgetLock);
	CList_emptyList((CListHandle *)&AttachedDevices, TRUE, CPhidget_free);
	CThread_mutex_unlock(&phidgetLock);
#endif
	DPRINT("Central Thread exiting");
	return EPHIDGET_OK;
}

//The read thread
CThread_func_return_t ReadThreadFunction(CThread_func_arg_t lpdwParam)
{
	CPhidgetHandle phid = (CPhidgetHandle)lpdwParam;
	int result;
	DPRINT("ReadThreadFunction running\n");

	if (!phid)
	{
		DPRINT("ReadThreadFunction returned 1\n");
		return (CThread_func_return_t)EPHIDGET_INVALIDARG;
	}
	while (phid->status == PHIDGET_ATTACHED)
		if((result=CPhidget_read(phid)))
		{
			DPRINT("ReadThreadFunction returned 2 : %d\n",result);
			return (CThread_func_return_t)result;
		}

	DPRINT("ReadThreadFunction returned 3\n");
	return (CThread_func_return_t)EPHIDGET_OK;
}

//The write thread
CThread_func_return_t WriteThreadFunction(CThread_func_arg_t lpdwParam)
{
	CPhidgetHandle phid = (CPhidgetHandle)lpdwParam;
	int result = 0;
	DPRINT("WriteThreadFunction running\n");

	if (!phid)
	{
		DPRINT("WriteThreadFunction returned 1\n");
		return (CThread_func_return_t)EPHIDGET_INVALIDARG;
	}


	while (phid->status == PHIDGET_ATTACHED) {
		switch (CThread_wait_on_event(&phid->writeAvailableEvent, 200)) {
		case WAIT_TIMEOUT:
			if(phid->writeStopFlag) 
			{
				phid->writeStopFlag = 0;
				DPRINT("WriteThreadFunction returned 2\n");
				return (CThread_func_return_t)EPHIDGET_OK;
			}
			if ((phid->awdc_enabled) && ((result = CPhidget_write(phid)) != EPHIDGET_OK))
			{
				DPRINT("WriteThreadFunction returned 3\n");
				return (CThread_func_return_t)result;  // if awdc_enabled is true, then we attempt to write.  If write fails, we bail out.
			}
			break;				
		case WAIT_OBJECT_0: //writeAvailable
			if ((result = CPhidget_write(phid)) != EPHIDGET_OK)
			{
				DPRINT("WriteThreadFunction returned 4\n");
				return (CThread_func_return_t)result;
			}
			break;
		default:
			DPRINT("WriteThreadFunction returned 5\n");
			return (CThread_func_return_t)EPHIDGET_UNEXPECTED;
		}
	}
	DPRINT("WriteThreadFunction returned 6\n");

	return (CThread_func_return_t)EPHIDGET_OK;
}

int
CThread_create(CThread *cp, CThread_func_t fp, CThread_func_arg_t arg)
{
#ifdef _WINDOWS
	//printf("Trying to create reading thread?\n");
	cp->m_ThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)fp, arg, 0, &cp->m_ThreadIdentifier);
	if(cp->m_ThreadHandle) return EPHIDGET_OK;
	else return GetLastError();
#else
	return pthread_create(&cp->m_ThreadHandle, NULL, fp, arg);
#endif
}

void
CThread_join(CThread *cp)
{
#ifdef _WINDOWS
	DWORD ec;

	while (GetExitCodeThread(cp->m_ThreadHandle, &ec) && ec == STILL_ACTIVE)
		SLEEP(100);
#else
	pthread_join(cp->m_ThreadHandle, 0);
#endif
}

int
CThread_mutex_init(CThread_mutex_t *mp)
{
#ifdef _WINDOWS
	InitializeCriticalSection(mp);
	return 1;
#else
	return pthread_mutex_init(mp, NULL) == 0;
#endif
}

int
CThread_mutex_destroy(CThread_mutex_t *mp)
{
#ifdef _WINDOWS
	DeleteCriticalSection(mp);
	return 1;
#else
	return pthread_mutex_destroy(mp) == 0;
#endif
}

void
CThread_mutex_lock(CThread_mutex_t *mp)
{
#ifdef _WINDOWS
	EnterCriticalSection(mp);
#else
	pthread_mutex_lock(mp);
#endif
}

void
CThread_mutex_unlock(CThread_mutex_t *mp)
{
#ifdef _WINDOWS
	LeaveCriticalSection(mp);
#else
	pthread_mutex_unlock(mp);
#endif
}

void CThread_create_event(EVENT *ev)
{
#ifdef _WINDOWS
	*ev = CreateEvent(NULL, FALSE, FALSE, NULL);
#else
    pthread_mutex_init(&ev->mutex, NULL);
    pthread_cond_init(&ev->condition, NULL);
	ev->ready_to_go = PFALSE;
#endif
}

int CThread_destroy_event(EVENT *ev)
{
#ifdef _WINDOWS
	return CloseHandle(*ev);
#else
    if(pthread_mutex_destroy(&ev->mutex)) return 0;
    if(pthread_cond_destroy(&ev->condition)) return 0;
	return 1;
#endif
}

int CThread_wait_on_event(EVENT *ev, EVENT_TIME time)
{
#ifdef _WINDOWS
	return WaitForSingleObject(*ev, time);
#else
	int retval;
	struct timespec timeout;
	struct timeval now;
	
    // Lock the mutex.
    pthread_mutex_lock(&ev->mutex);
 
    // If the predicate is already set, then the while loop is bypassed;
    // otherwise, the thread sleeps until the predicate is set.
    if(ev->ready_to_go == PFALSE)
    {
		if(time == INFINITE)
			retval = pthread_cond_wait(&ev->condition, &ev->mutex);
		else {
			gettimeofday(&now,0);
			timeout.tv_sec = now.tv_sec;
			timeout.tv_nsec = now.tv_usec*1000 + (time * 1000000);
			retval = pthread_cond_timedwait(&ev->condition, &ev->mutex, &timeout);
		}
		
		switch(retval)
		{
			case ETIMEDOUT:
				pthread_mutex_unlock(&ev->mutex); 
				return WAIT_TIMEOUT;
			//case EINVAL:
			//	return WAIT_FAILED;
			default:
				break;
		}
    }
	pthread_mutex_unlock(&ev->mutex); 
	return WAIT_OBJECT_0;
#endif
}

void CThread_reset_event(EVENT *ev)
{
#ifdef _WINDOWS
	ResetEvent(*ev);
#else
    // Reset the predicate and release the mutex.
    pthread_mutex_lock(&ev->mutex);
    ev->ready_to_go = PFALSE;
    pthread_mutex_unlock(&ev->mutex); 
#endif
}

void CThread_set_event(EVENT *ev)
{
#ifdef _WINDOWS
	SetEvent(*ev);
#else
    // At this point, there should be work for the other thread to do.
    pthread_mutex_lock(&ev->mutex);
    ev->ready_to_go = PTRUE;
    // Signal the other thread to begin work.
    pthread_cond_signal(&ev->condition);
    pthread_mutex_unlock(&ev->mutex);
 
#endif
}

